"""Implements the two primary data containers for correlation function data,
:obj:`CorrFunc` which stores the pair counts and :obj:`CorrData` which stores
the (resampled) values of the correlation function in bins of redshift.
"""

from __future__ import annotations

import logging
from collections.abc import Sequence
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Any, Type, TypeVar

import h5py
import numpy as np
import pandas as pd
from deprecated import deprecated

from yaw.catalogs import PatchLinkage
from yaw.config import OPTIONS, ResamplingConfig
from yaw.core.abc import BinnedQuantity, HDFSerializable, PatchedQuantity
from yaw.core.containers import Indexer, SampledData
from yaw.core.logging import LogCustomWarning, TimedLog
from yaw.core.utils import TypePathStr
from yaw.core.utils import format_float_fixed_width as fmt_num
from yaw.correlation.estimators import (
    CorrelationEstimator,
    CtsMix,
    EstimatorError,
    cts_from_code,
)
from yaw.correlation.paircounts import NormalisedCounts, TypeIndex

if TYPE_CHECKING:  # pragma: no cover
    from matplotlib.axis import Axis
    from numpy.typing import NDArray
    from pandas import IntervalIndex

    from yaw.catalogs import BaseCatalog
    from yaw.config import Configuration
    from yaw.correlation.estimators import Cts

__all__ = ["CorrData", "CorrFunc", "add_corrfuncs", "autocorrelate", "crosscorrelate"]


logger = logging.getLogger(__name__)

_Tdata = TypeVar("_Tdata", bound="CorrData")


@dataclass(frozen=True, repr=False, eq=False)
class CorrData(SampledData):
    """Container class for sampled correlation function data.

    Contains the redshift binning, correlation function amplitudes, and
    resampled amplitudes (e.g. jackknife or bootstrap). The resampled values are
    used to compute error estimates and covariance/correlation matrices.
    Provides some plotting methods for convenience.

    The comparison, addition and subtraction and indexing rules are inherited
    from :obj:`~yaw.core.containers.SampledData`, see some examples below.

    .. rubric:: Examples

    Create a new instance by sampling a correlation function:

    >>> from yaw.examples import w_sp
    >>> data = w_sp.sample()  # uses the default ResamplingConfig
    >>> data
    CorrData(n_bins=30, z='0.070...1.420', n_samples=64, method='jackknife')

    View the data for a subset of the redshift bins:

    >>> data.bins[5:9].data
    array([0.10158809, 0.08079947, 0.03876175, 0.02715336])

    View the same subset as series:

    >>> data.bins[5:9].get_data()
    (0.295, 0.34]    0.101588
    (0.34, 0.385]    0.080799
    (0.385, 0.43]    0.038762
    (0.43, 0.475]    0.027153
    dtype: float64

    Get the redshift bin centers for these bins:

    >>> data.bins[5:9].mids
    array([0.3175, 0.3625, 0.4075, 0.4525])

    Args:
        binning (:obj:`pandas.IntervalIndex`):
            The redshift bin edges used for this correlation function.
        data (:obj:`NDArray`):
            The correlation function values.
        samples (:obj:`NDArray`):
            The resampled correlation function values.
        method (:obj:`str`):
            The resampling method used, see :class:`~yaw.ResamplingConfig` for
            available options.
        info (:obj:`str`, optional):
            Descriptive text included in the headers of output files produced
            by :func:`CorrData.to_files`.
    """

    info: str | None = None
    """Optional descriptive text for the contained data."""

    def __post_init__(self) -> None:
        with LogCustomWarning(
            logger, "invalid values encountered in correlation samples"
        ):
            super().__post_init__()

    @classmethod
    def from_files(cls: Type[_Tdata], path_prefix: TypePathStr) -> _Tdata:
        """Create a new instance by loading the data from ASCII files.

        The data is restored from a set of three input files produced by
        :meth:`to_files`.

        .. Note::
            These file have the same names but different file extension,
            therefore only provide the base name without any extension to
            specifiy the input files.

        Args:
            path_prefix (:obj:`str`):
                The base name of the input files without any file extension.

        Returns:
            :obj:`CorrData`
        """
        name = cls.__name__.lower()[:-4]
        logger.debug(f"reading {name} data from '{path_prefix}.*'")
        # load data and errors
        ext = "dat"
        data_error = np.loadtxt(f"{path_prefix}.{ext}")
        # restore index
        binning = pd.IntervalIndex.from_arrays(data_error[:, 0], data_error[:, 1])
        # load samples
        ext = "smp"
        samples = np.loadtxt(f"{path_prefix}.{ext}")
        # load header
        info = None
        with open(f"{path_prefix}.{ext}") as f:
            for line in f.readlines():
                if "extra info" in line:
                    _, info = line.split(":", maxsplit=1)
                    info = info.strip()
                if "z_low" in line:
                    line = line[2:].strip("\n")  # remove leading '# '
                    header = [col for col in line.split(" ") if len(col) > 0]
                    break
            else:
                raise ValueError("sample file header misformatted")
        method_key, n_samples = header[-1].rsplit("_", 1)
        n_samples = int(n_samples) + 1
        # reconstruct sampling method
        for method in OPTIONS.method:
            if method.startswith(method_key):
                break
        else:
            raise ValueError(f"invalid sampling method key '{method_key}'")
        return cls(
            binning=binning,
            data=data_error[:, 2],  # take data column
            samples=samples.T[2:],  # remove redshift bin columns
            method=method,
            info=info,
        )

    @property
    def _dat_desc(self) -> str:
        """Description included in the data file."""
        return (
            "# correlation function estimate with symmetric 68% percentile "
            "confidence"
        )

    @property
    def _smp_desc(self) -> str:
        """Description included in the samples file."""
        return f"# {self.n_samples} {self.method} correlation function samples"

    @property
    def _cov_desc(self) -> str:
        """Description included in the covariance file."""
        return (
            f"# correlation function estimate covariance matrix "
            f"({self.n_bins}x{self.n_bins})"
        )

    def to_files(self, path_prefix: TypePathStr) -> None:
        """Store the data in a set of ASCII files on disk.

        These files can be loaded with the :meth:`from_files` method. There are
        three files with the same name but different file extension.

        .. rubric:: Files

        ``[path_prefix].dat``: Contains the redshift bin edges, the data values
        and their standard error. Additionally there is information about the
        error estimate and the :obj:`info` attribute.

        ``[path_prefix].smp``: Contains one row for each redshift bin. The first
        two columns list the lower and upper edge of the redshift bin, the
        remaining columns list the values of the samples, i.e. there are ``N+2``
        columns. Additionally contains the :obj:`info` attribute.

        ``[path_prefix].cov``: Contains the covariance matrix and additionally
        the :obj:`info` attribute.

        Args:
            path_prefix (:obj:`str`):
                The base name of the output files without any file extension.
        """
        name = self.__class__.__name__.lower()[:-4]
        logger.info(f"writing {name} data to '{path_prefix}.*'")
        PREC = 10
        DELIM = " "

        def comment(string: str) -> str:
            if self.info is not None:
                string = f"{string}\n# extra info: {self.info}"
            return string

        def write_head(f, description, header, delim=DELIM):
            f.write(f"{description}\n")
            line = delim.join(f"{h:>{PREC}s}" for h in header)
            f.write(f"# {line[2:]}\n")

        # write data and errors
        ext = "dat"
        header = ["z_low", "z_high", "nz", "nz_err"]
        with open(f"{path_prefix}.{ext}", "w") as f:
            write_head(f, comment(self._dat_desc), header, delim=DELIM)
            for zlow, zhigh, nz, nz_err in zip(
                self.edges[:-1], self.edges[1:], self.data, self.error
            ):
                values = [fmt_num(val, PREC) for val in (zlow, zhigh, nz, nz_err)]
                f.write(DELIM.join(values) + "\n")

        # write samples
        ext = "smp"
        header = ["z_low", "z_high"]
        header.extend(f"{self.method[:4]}_{i}" for i in range(self.n_samples))
        with open(f"{path_prefix}.{ext}", "w") as f:
            write_head(f, comment(self._smp_desc), header, delim=DELIM)
            for zlow, zhigh, samples in zip(
                self.edges[:-1], self.edges[1:], self.samples.T
            ):
                values = [fmt_num(zlow, PREC), fmt_num(zhigh, PREC)]
                values.extend(fmt_num(val, PREC) for val in samples)
                f.write(DELIM.join(values) + "\n")

        # write covariance (just for convenience)
        ext = "cov"
        fmt_str = DELIM.join("{: .{prec}e}" for _ in range(self.n_bins)) + "\n"
        with open(f"{path_prefix}.{ext}", "w") as f:
            f.write(f"{comment(self._cov_desc)}\n")
            for values in self.covariance:
                f.write(fmt_str.format(*values, prec=PREC - 3))

    def _make_plot(
        self,
        x: NDArray[np.float_],
        y: NDArray[np.float_],
        yerr: NDArray[np.float_],
        *,
        color: str | NDArray | None = None,
        label: str | None = None,
        error_bars: bool = True,
        ax: Axis | None = None,
        plot_kwargs: dict[str, Any] | None = None,
        zero_line: bool = False,
    ) -> Axis:
        from matplotlib import pyplot as plt

        # configure plot
        if ax is None:
            ax = plt.gca()
        if plot_kwargs is None:
            plot_kwargs = {}
        plot_kwargs.update(dict(color=color, label=label))
        ebar_kwargs = dict(fmt=".", ls="none")
        ebar_kwargs.update(plot_kwargs)
        # plot zero line
        if zero_line:
            lw = 0.7
            for spine in ax.spines.values():
                lw = spine.get_linewidth()
            ax.axhline(0.0, color="k", lw=lw, zorder=-2)
        # plot data
        if error_bars:
            ax.errorbar(x, y, yerr, **ebar_kwargs)
        else:
            color = ax.plot(x, y, **plot_kwargs)[0].get_color()
            ax.fill_between(x, y - yerr, y + yerr, color=color, alpha=0.2)
        return ax

    def plot(
        self,
        *,
        color: str | NDArray | None = None,
        label: str | None = None,
        error_bars: bool = True,
        ax: Axis | None = None,
        xoffset: float = 0.0,
        plot_kwargs: dict[str, Any] | None = None,
        zero_line: bool = False,
        scale_by_dz: bool = False,
    ) -> Axis:  # pragma: no cover
        """Create a plot of the correlation data as a function of redshift.

        Create a new axis or plot to an existing one, add x-axis offsets, if
        plotting multiple instances, or specify if the values should be
        represented as points with errorbars (default) or as line plot with
        shaded area to represent uncertainties.

        Args:
            color:
                Valid :mod:`matplotlib` color used for the error bars or the
                line and the shaded uncertainty area.
            label (:obj:`str`, optional):
                Plot label for the legend.
            error_bars (:obj:`bool`, optional):
                Whether to plot error bars (the default) or a line plot with
                shaded area.
            ax (plot axis, optional):
                Optional :mod:`matplotlib` axis to plot into.
            xoffset (:obj:`int`, optional):
                Shift to apply to the x-axis (redshift) values.
            plot_kwargs (:obj:`dict`, optional):
                Parameters passed to the :func:`errobar` or :func:`plot`
                plotting functions.
            zero_lilne (:obj:`bool`, optional):
                Wether to draw a thin black line that indicates ``y=0``.
            scale_by_dz (:obj:`bool`, optional):
                Whether to multiply the y-values by the redshift bin width
                :obj:`dz`.
        """
        x = self.mids + xoffset
        y = self.data
        yerr = self.error
        if scale_by_dz:
            y *= self.dz
            yerr *= self.dz
        return self._make_plot(
            x,
            y,
            yerr,
            color=color,
            label=label,
            error_bars=error_bars,
            ax=ax,
            plot_kwargs=plot_kwargs,
            zero_line=zero_line,
        )

    def plot_corr(
        self, *, redshift: bool = False, cmap: str = "RdBu_r", ax: Axis | None = None
    ) -> Axis:
        """Plot the correlation matrix of the data.

        Create a new axis or plot to an existing one.

        Args:
            redshift (:obj:`bool`, optional):
                Whether to map the matrix onto redshifts or as regular matrix
                plot (the default).
            cmap (:obj:`str`, optional):
                Name of a :mod:`matplotlib` colormap to use.
            ax (plot axis, optional):
                Optional :mod:`matplotlib` axis to plot into.
        """
        from matplotlib import pyplot as plt

        if ax is None:
            ax = plt.gca()
        corr = self.get_correlation()
        cmap_kwargs = dict(cmap=cmap, vmin=-1.0, vmax=1.0)
        if redshift:
            ticks = self.mids
            ax.pcolormesh(ticks, ticks, np.flipud(corr), **cmap_kwargs)
            ax.xaxis.tick_top()
            ax.set_aspect("equal")
        else:
            ax.matshow(corr, **cmap_kwargs)
        return ax


def check_mergable(cfs: Sequence[CorrFunc | None]) -> None:
    """Helper function that checks if a set of :obj:`CorrFunc` have the same
    kinds of pair counts."""
    reference = cfs[0]
    for kind in ("dd", "dr", "rd", "rr"):
        ref_pcounts = getattr(reference, kind)
        for cf in cfs[1:]:
            pcounts = getattr(cf, kind)
            if type(ref_pcounts) != type(pcounts):
                raise ValueError(f"cannot merge, '{kind}' incompatible")


@dataclass(frozen=True)
class CorrFunc(PatchedQuantity, BinnedQuantity, HDFSerializable):
    """Container object for measured correlation pair counts.

    Container returned by :meth:`~yaw.catalogs.BaseCatalog.correlate` that
    computes the correlations between data catalogs. The correlation function
    can be computed from four kinds of pair counts, data-data (DD), data-random
    (DR), random-data (RD), and random-random (RR).

    .. Note::
        DD is always required, but DR, RD, and RR are optional as long as at
        least one is provided.

    Provides methods to read and write data to disk and compute the actual
    correlation function values (see :class:`~yaw.CorrData`) using spatial
    resampling (see :class:`~yaw.ResamplingConfig`).

    The container supports comparison with ``==`` and ``!=`` on the pair count
    level. The supported arithmetic operations between two correlation
    functions, addition and subtraction, are applied between all internally
    stored pair counts data. The same applies to rescaling of the counts by a
    scalar, see some examples below.

    .. rubric:: Examples

    Create a new instance by sampling a correlation function:

    >>> from yaw.examples import w_sp
    >>> dd, dr = w_sp.dd, w_sp.dr  # get example data-data and data-rand counts
    >>> corr = yaw.CorrFunc(dd=dd, dr=dr)
    >>> corr
    CorrFunc(n_bins=30, z='0.070...1.420', dd=True, dr=True, rd=False, rr=False, n_patches=64)

    Access the pair counts:

    >>> corr.dd
    NormalisedCounts(n_bins=30, z='0.070...1.420', n_patches=64)

    Check if it is an autocorrelation function measurement:

    >>> corr.auto
    False

    Check which pair counts are available to compute the correlation function:

    >>> corr.estimators
    {'DP': yaw.correlation.estimators.DavisPeebles}

    Sample the correlation function

    >>> corr.sample()  # uses the default ResamplingConfig
    CorrData(n_bins=30, z='0.070...1.420', n_samples=64, method='jackknife')

    Note how the indicated shape changes when a patch subset is selected:

    >>> corr.patches[:10]
    CorrFunc(n_bins=30, z='0.070...1.420', dd=True, dr=True, rd=False, rr=False, n_patches=10)

    Note how the indicated redshift range and shape change when a bin subset is
    selected:

    >>> corr.bins[:3]
    CorrFunc(n_bins=3, z='0.070...0.205', dd=True, dr=True, rd=False, rr=False, n_patches=64)

    Args:
        dd (:obj:`~yaw.correlation.paircounts.NormalisedCounts`):
            Pair counts from a data-data count measurement.
        dr (:obj:`~yaw.correlation.paircounts.NormalisedCounts`, optional):
            Pair counts from a data-random count measurement.
        rd (:obj:`~yaw.correlation.paircounts.NormalisedCounts`, optional):
            Pair counts from a random-data count measurement.
        rr (:obj:`~yaw.correlation.paircounts.NormalisedCounts`, optional):
            Pair counts from a random-random count measurement.
    """

    dd: NormalisedCounts
    """Pair counts for a data-data correlation measurement"""
    dr: NormalisedCounts | None = field(default=None)
    """Pair counts from a data-random count measurement."""
    rd: NormalisedCounts | None = field(default=None)
    """Pair counts from a random-data count measurement."""
    rr: NormalisedCounts | None = field(default=None)
    """Pair counts from a random-random count measurement."""

    def __post_init__(self) -> None:
        # check if any random pairs are required
        if self.dr is None and self.rd is None and self.rr is None:
            raise ValueError("either 'dr', 'rd' or 'rr' is required")
        # check that the pair counts are compatible
        for kind in ("dr", "rd", "rr"):
            pairs: NormalisedCounts | None = getattr(self, kind)
            if pairs is None:
                continue
            try:
                self.dd.is_compatible(pairs, require=True)
            except ValueError as e:
                raise ValueError(
                    f"pair counts '{kind}' and 'dd' are not compatible"
                ) from e

    def __repr__(self) -> str:
        string = super().__repr__()[:-1]
        pairs = f"dd=True, dr={self.dr is not None}, "
        pairs += f"rd={self.rd is not None}, rr={self.rr is not None}"
        other = f"n_patches={self.n_patches}"
        return f"{string}, {pairs}, {other})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        for cfield in fields(self):
            kind = cfield.name
            if getattr(self, kind) != getattr(other, kind):
                return False
        return True

    def __neq__(self, other) -> bool:
        return not self == other

    def __add__(self, other: CorrFunc) -> CorrFunc:
        # check that the pair counts are set consistently
        kinds = []
        for cfield in fields(self):
            kind = cfield.name
            self_set = getattr(self, kind) is not None
            other_set = getattr(other, kind) is not None
            if (self_set and not other_set) or (not self_set and other_set):
                raise ValueError(f"pair counts for '{kind}' not set for both operands")
            elif self_set and other_set:
                kinds.append(kind)

        kwargs = {kind: getattr(self, kind) + getattr(other, kind) for kind in kinds}
        return self.__class__(**kwargs)

    def __radd__(self, other: CorrFunc | int | float) -> CorrFunc:
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __mul__(self, other: np.number) -> CorrFunc:
        # check that the pair counts are set consistently
        kwargs = {}
        for cfield in fields(self):
            kind = cfield.name
            counts = getattr(self, kind)
            if counts is not None:
                kwargs[kind] = counts * other
        return self.__class__(**kwargs)

    @property
    def auto(self) -> bool:
        """Whether the stored data are from an autocorrelation measurement."""
        return self.dd.auto

    @property
    def bins(self) -> Indexer[TypeIndex, CorrFunc]:
        def builder(inst: CorrFunc, item: TypeIndex) -> CorrFunc:
            if isinstance(item, int):
                item = [item]
            kwargs = {}
            for cfield in fields(inst):
                pairs: NormalisedCounts | None = getattr(inst, cfield.name)
                if pairs is None:
                    kwargs[cfield.name] = None
                else:
                    kwargs[cfield.name] = pairs.bins[item]
            return CorrFunc(**kwargs)

        return Indexer(self, builder)

    @property
    def patches(self) -> Indexer[TypeIndex, CorrFunc]:
        def builder(inst: CorrFunc, item: TypeIndex) -> CorrFunc:
            kwargs = {}
            for cfield in fields(inst):
                counts: NormalisedCounts | None = getattr(inst, cfield.name)
                if counts is not None:
                    counts = counts.patches[item]
                kwargs[cfield.name] = counts
            return CorrFunc(**kwargs)

        return Indexer(self, builder)

    def get_binning(self) -> IntervalIndex:
        return self.dd.get_binning()

    @property
    def n_patches(self) -> int:
        return self.dd.n_patches

    def is_compatible(self, other: CorrFunc, require: bool = True) -> bool:
        """Check whether this instance is compatible with another instance.

        Ensures that the redshift binning and the number of patches are
        identical.

        Args:
            other (:obj:`BinnedQuantity`):
                Object instance to compare to.
            require (:obj:`bool`)
                Raise a ValueError if any of the checks fail.

        Returns:
            :obj:`bool`
        """
        if self.dd.n_patches != other.dd.n_patches:
            if require:
                raise ValueError("number of patches does not agree")
            return False
        return self.dd.is_compatible(other.dd, require)

    @property
    def estimators(self) -> dict[str, CorrelationEstimator]:
        """Get a listing of correlation estimators implemented, depending on
        which pair counts are available.

        Returns:
            :obj:`dict`: Mapping from correlation estimator name abbreviation to
            correlation function class.
        """
        # figure out which of dd, dr, ... are not None
        available = set()
        # iterate all dataclass attributes that are in __init__
        for attr in fields(self):
            if not attr.init:
                continue
            if getattr(self, attr.name) is not None:
                available.add(cts_from_code(attr.name))
        # check which estimators are supported
        estimators = {}
        for estimator in CorrelationEstimator.variants:  # registered estimators
            if set(estimator.requires) <= available:
                estimators[estimator.short] = estimator
        return estimators

    def _check_and_select_estimator(
        self, estimator: str | None
    ) -> type[CorrelationEstimator]:
        options = self.estimators
        if estimator is None:
            for shortname in ["LS", "DP", "PH"]:  # preferred hierarchy
                if shortname in options:
                    estimator = shortname
                    break
        estimator = estimator.upper()
        if estimator not in options:
            try:
                index = [e.short for e in CorrelationEstimator.variants].index(
                    estimator
                )
                est_class = CorrelationEstimator.variants[index]
            except ValueError as e:
                raise ValueError(f"invalid estimator '{estimator}'") from e
            # determine which pair counts are missing
            for attr in fields(self):
                name = attr.name
                if not attr.init:
                    continue
                cts = cts_from_code(name)
                if getattr(self, name) is None and cts in est_class.requires:
                    raise EstimatorError(f"estimator requires {name}")
            else:
                raise RuntimeError()
        # select the correct estimator
        cls = options[estimator]
        logger.debug(
            f"selecting estimator '{cls.short}' from {'/'.join(self.estimators)}"
        )
        return cls

    def _getattr_from_cts(self, cts: Cts) -> NormalisedCounts | None:
        if isinstance(cts, CtsMix):
            for code in str(cts).split("_"):
                value = getattr(self, code)
                if value is not None:
                    break
            return value
        else:
            return getattr(self, str(cts))

    @deprecated(reason="renamed to CorrFunc.sample", version="2.3.1")
    def get(self, *args, **kwargs):
        """
        .. deprecated:: 2.3.1
            Renamed to :meth:`sample`.
        """
        return self.sample(*args, **kwargs)

    def sample(
        self,
        config: ResamplingConfig | None = None,
        *,
        estimator: str | None = None,
        info: str | None = None,
    ) -> CorrData:
        """Compute the correlation function from the stored pair counts,
        including an error estimate from spatial resampling of patches.

        Args:
            config (:obj:`~yaw.ResamplingConfig`):
                Specify the resampling method and its configuration.

        Keyword Args:
            estimator (:obj:`str`, optional):
                The name abbreviation for the correlation estimator to use.
                Defaults to Landy-Szalay if RR is available, otherwise to
                Davis-Peebles.
            info (:obj:`str`, optional):
                Descriptive text passed on to the output :obj:`CorrData`
                object.

        Returns:
            :obj:`CorrData`:
                Correlation function data, including redshift binning, function
                values and samples.
        """
        if config is None:
            config = ResamplingConfig()
        est_fun = self._check_and_select_estimator(estimator)
        logger.debug(f"computing correlation and {config.method} samples")
        # get the pair counts for the required terms (DD, maybe DR and/or RR)
        required_data = {}
        required_samples = {}
        for cts in est_fun.requires:
            try:  # if pairs are None, estimator with throw error
                pairs = self._getattr_from_cts(cts).sample(config)
                required_data[str(cts)] = pairs.data
                required_samples[str(cts)] = pairs.samples
            except AttributeError as e:
                if "NoneType" not in e.args[0]:
                    raise
        # get the pair counts for the optional terms (e.g. RD)
        optional_data = {}
        optional_samples = {}
        for cts in est_fun.optional:
            try:  # if pairs are None, estimator with throw error
                pairs = self._getattr_from_cts(cts).sample(config)
                optional_data[str(cts)] = pairs.data
                optional_samples[str(cts)] = pairs.samples
            except AttributeError as e:
                if "NoneType" not in e.args[0]:
                    raise
        # evaluate the correlation estimator
        data = est_fun.eval(**required_data, **optional_data)
        samples = est_fun.eval(**required_samples, **optional_samples)
        return CorrData(
            binning=self.get_binning(),
            data=data,
            samples=samples,
            method=config.method,
            info=info,
        )

    @classmethod
    def from_hdf(cls, source: h5py.File | h5py.Group) -> CorrFunc:
        def _try_load(root: h5py.Group, name: str) -> NormalisedCounts | None:
            try:
                return NormalisedCounts.from_hdf(root[name])
            except KeyError:
                return None

        dd = NormalisedCounts.from_hdf(source["data_data"])
        dr = _try_load(source, "data_random")
        rd = _try_load(source, "random_data")
        rr = _try_load(source, "random_random")
        return cls(dd=dd, dr=dr, rd=rd, rr=rr)

    def to_hdf(self, dest: h5py.File | h5py.Group) -> None:
        group = dest.create_group("data_data")
        self.dd.to_hdf(group)
        group_names = dict(dr="data_random", rd="random_data", rr="random_random")
        for kind, name in group_names.items():
            data: NormalisedCounts | None = getattr(self, kind)
            if data is not None:
                group = dest.create_group(name)
                data.to_hdf(group)
        dest.create_dataset("n_patches", data=self.n_patches)

    @classmethod
    def from_file(cls, path: TypePathStr) -> CorrFunc:
        logger.debug(f"reading pair counts from '{path}'")
        with h5py.File(str(path)) as f:
            return cls.from_hdf(f)

    def to_file(self, path: TypePathStr) -> None:
        logger.info(f"writing pair counts to '{path}'")
        with h5py.File(str(path), mode="w") as f:
            self.to_hdf(f)

    def concatenate_patches(self, *cfs: CorrFunc) -> CorrFunc:
        check_mergable([self, *cfs])
        merged = {}
        for kind in ("dd", "dr", "rd", "rr"):
            self_pcounts = getattr(self, kind)
            if self_pcounts is not None:
                other_pcounts = [getattr(cf, kind) for cf in cfs]
                merged[kind] = self_pcounts.concatenate_patches(*other_pcounts)
        return self.__class__(**merged)

    def concatenate_bins(self, *cfs: CorrFunc) -> CorrFunc:
        check_mergable([self, *cfs])
        merged = {}
        for kind in ("dd", "dr", "rd", "rr"):
            self_pcounts = getattr(self, kind)
            if self_pcounts is not None:
                other_pcounts = [getattr(cf, kind) for cf in cfs]
                merged[kind] = self_pcounts.concatenate_bins(*other_pcounts)
        return self.__class__(**merged)


def _create_dummy_counts(counts: Any | dict[str, Any]) -> dict[str, None]:
    """Duplicate a the return values of
    :meth:`yaw.catalogs.BaseCatalog.correlate`, but replace the :obj:`CorrFunc`
    instances by :obj:`None`."""
    if isinstance(counts, dict):
        dummy = {scale_key: None for scale_key in counts}
    else:
        dummy = None
    return dummy


def add_corrfuncs(
    corrfuncs: Sequence[CorrFunc], weights: Sequence[np.number] | None = None
) -> CorrFunc:
    """Add correlation functions that are measured at different scales.

    The correlation functions are added by summing together their pair counts.
    They can be weighted prior to summation by effectively scaling their pair
    counts with a set of scalar weights, one for each input correlation
    function.

    .. Note::
        The actual scales are not checked, but the number of patches and the
        redshift binning of the inputs must be identical.

    This operation is effectively equivalent to:

    >>> corrfunc1 * weight1 + corrfunc2 * weight2  # + ...

    Args:
        corrfuncs (sequence of :obj:`CorrFunc`):
            A list of correlation functions to add.
        weights (sequence of :obj:`int` or :obj:`float`, optional):
            An optional list of weights, one for each correlation function.

    Returns:
        :obj:`CorrFunc`:
            The combined correlation function after summing the pairs.
    """
    if weights is None:
        weights = [1.0] * len(corrfuncs)
    else:
        if len(corrfuncs) != len(weights):
            raise ValueError(
                "number of weights must match number of correlation functions"
            )
    # run summation, rescaling by weights
    combined = 0.0
    for corrfunc, weight in zip(corrfuncs, weights):
        combined = combined + (corrfunc * weight)
    return combined


class PatchError(Exception):
    pass


def _check_patch_centers(catalogues: Sequence[BaseCatalog]) -> None:
    """Check whether the patch centers of a set of data catalogues are seperated
    by no more than the radius of the patches."""
    refcat = catalogues[0]
    for cat in catalogues[1:]:
        if refcat.n_patches != cat.n_patches:
            raise PatchError("number of patches does not agree")
        ref_coord = refcat.centers.to_sky()
        cat_coord = cat.centers.to_sky()
        dist = ref_coord.distance(cat_coord)
        if np.any(dist.values > refcat.radii.values):
            raise PatchError("the patch centers are inconsistent")


def autocorrelate(
    config: Configuration,
    data: BaseCatalog,
    random: BaseCatalog,
    *,
    linkage: PatchLinkage | None = None,
    compute_rr: bool = True,
    progress: bool = False,
) -> CorrFunc | dict[str, CorrFunc]:
    """Compute an angular autocorrelation function in bins of redshift.

    The correlation is measured on fixed physical scales that are converted to
    angles for each redshift bin. All parameters (binning, scales, etc.) are
    bundled in the input configuration, see :mod:`yaw.config`.

    .. Note::
        Both the data and random catalogue require redshift point estimates.

    Args:
        config (:obj:`~yaw.config.Configuration`):
            Provides all major run parameters, such as scales, binning, and for
            the correlation measurement backend.
        data (:obj:`~yaw.catalogs.BaseCatalog`):
            The data sample catalogue.
        random (:obj:`~yaw.catalogs.BaseCatalog`):
            Random catalogue for the data sample.

    Keyword Args:
        linkage (:obj:`~yaw.catalogs.PatchLinkage`, optional):
            Provide a linkage object that determines which spatial patches must
            be correlated given the measurement scales. Ensures consistency
            when measuring correlations repeatedly for a fixed set of input
            catalogues. Generated automatically by default.
        compute_rr (:obj:`bool`):
            Whether the random-random (RR) pair counts are computed.
        progress (:obj:`bool`):
            Display a progress bar.

    Returns:
        :obj:`CorrFunc` or :obj:`dict[str, CorrFunc]`:
            Container that holds the measured pair counts, or a dictionary of
            containers if multiple scales are configured. Dictionary keys have a
            ``kpcXXtXX`` pattern, where ``XX`` are the lower and upper scale
            limit as integers, in kpc (see :obj:`yaw.core.cosmology.Scale`).
    """
    _check_patch_centers([data, random])
    scales = config.scales.as_array()
    logger.info(
        f"running autocorrelation ({len(scales)} scales, "
        f"{scales.min():.0f}<r<={scales.max():.0f}kpc)"
    )
    if linkage is None:
        linkage = PatchLinkage.from_setup(config, random)
    kwargs = dict(linkage=linkage, progress=progress)
    logger.debug("scheduling DD, DR" + (", RR" if compute_rr else ""))
    with TimedLog(logger.info, "counting data-data pairs"):
        DD = data.correlate(config, binned=True, **kwargs)
    with TimedLog(logger.info, "counting data-rand pairs"):
        DR = data.correlate(config, binned=True, other=random, **kwargs)
    if compute_rr:
        with TimedLog(logger.info, "counting rand-rand pairs"):
            RR = random.correlate(config, binned=True, **kwargs)
    else:
        RR = _create_dummy_counts(DD)

    if isinstance(DD, dict):
        result = {
            scale: CorrFunc(dd=DD[scale], dr=DR[scale], rr=RR[scale]) for scale in DD
        }
    else:
        result = CorrFunc(dd=DD, dr=DR, rr=RR)
    return result


def crosscorrelate(
    config: Configuration,
    reference: BaseCatalog,
    unknown: BaseCatalog,
    *,
    ref_rand: BaseCatalog | None = None,
    unk_rand: BaseCatalog | None = None,
    linkage: PatchLinkage | None = None,
    progress: bool = False,
) -> CorrFunc | dict[str, CorrFunc]:
    """Compute an angular crosscorrelation function in bins of redshift.

    The correlation is measured on fixed physical scales that are converted to
    angles for each redshift bin. All parameters (binning, scales, etc.) are
    bundled in the input configuration, see :mod:`yaw.config`.

    At least one random catalogue (either for the reference or the unknown
    sample) must be provided, which will either trigger counting the DR
    (reference-random) or RD (random-unknown) pair counts. If both random
    catalogues are provided, the random-random pairs (RR) are counted as well,
    this is equivalent to enabling the ``compute_rr`` parameter in
    :func:`autocorrelate`.

    .. Note::
        The reference catalogue requires redshift point estimates. If the
        reference random cataloge is provided, it also requires redshifts.

    Args:
        config (:obj:`~yaw.config.Configuration`):
            Provides all major run parameters.
        reference (:obj:`yaw.catalogs.BaseCatalog`):
            The reference sample.
        unknown (:obj:`yaw.catalogs.BaseCatalog`):
            The sample with unknown redshift distribution.

    Keyword Args:
        ref_rand (:obj:`yaw.catalogs.BaseCatalog`, optional):
            Random catalog for the reference sample, requires redshifts
            configured.
        unk_rand (:obj:`yaw.catalogs.BaseCatalog`, optional):
            Random catalog for the unknown sample.
        linkage (:obj:`yaw.catalogs.PatchLinkage`, optional):
            Provide a linkage object that determines which spatial patches must
            be correlated given the measurement scales. Ensures consistency
            when measuring multiple correlations, otherwise generated
            automatically.
        progress (:obj:`bool`):
            Display a progress bar.

    Returns:
        :obj:`CorrFunc` or :obj:`dict[str, CorrFunc]`:
            Container that holds the measured pair counts, or a dictionary of
            containers if multiple scales are configured. Dictionary keys have a
            ``kpcXXtXX`` pattern, where ``XX`` are the lower and upper scale
            limit as integers, in kpc (see :obj:`yaw.core.cosmology.Scale`).
    """
    compute_dr = unk_rand is not None
    compute_rd = ref_rand is not None
    compute_rr = compute_dr and compute_rd
    # make sure that the patch centers are consistent
    all_cats = [reference, unknown]
    if compute_dr:
        all_cats.append(unk_rand)
    if compute_rd:
        all_cats.append(ref_rand)
    _check_patch_centers(all_cats)

    scales = config.scales.as_array()
    logger.info(
        f"running crosscorrelation ({len(scales)} scales, "
        f"{scales.min():.0f}<r<={scales.max():.0f}kpc)"
    )
    if linkage is None:
        linkage = PatchLinkage.from_setup(config, unknown)
    logger.debug(
        "scheduling DD"
        + (", DR" if compute_dr else "")
        + (", RD" if compute_rd else "")
        + (", RR" if compute_rr else "")
    )
    kwargs = dict(linkage=linkage, progress=progress)
    with TimedLog(logger.info, "counting data-data pairs"):
        DD = reference.correlate(config, binned=False, other=unknown, **kwargs)
    if compute_dr:
        with TimedLog(logger.info, "counting data-rand pairs"):
            DR = reference.correlate(config, binned=False, other=unk_rand, **kwargs)
    else:
        DR = _create_dummy_counts(DD)
    if compute_rd:
        with TimedLog(logger.info, "counting rand-data pairs"):
            RD = ref_rand.correlate(config, binned=False, other=unknown, **kwargs)
    else:
        RD = _create_dummy_counts(DD)
    if compute_rr:
        with TimedLog(logger.info, "counting rand-rand pairs"):
            RR = ref_rand.correlate(config, binned=False, other=unk_rand, **kwargs)
    else:
        RR = _create_dummy_counts(DD)

    if isinstance(DD, dict):
        result = {
            scale: CorrFunc(dd=DD[scale], dr=DR[scale], rd=RD[scale], rr=RR[scale])
            for scale in DD
        }
    else:
        result = CorrFunc(dd=DD, dr=DR, rd=RD, rr=RR)
    return result



"""This module implements the ``scipy`` backend for the Catalog class."""

from yaw.catalogs.scipy.catalog import ScipyCatalog
from yaw.catalogs.scipy.kdtree import SphericalKDTree
from yaw.catalogs.scipy.patches import PatchCatalog

__all__ = ["ScipyCatalog", "SphericalKDTree", "PatchCatalog"]



import numpy as np
from setuptools import Extension, setup

ext_module_math = Extension(
    "yaw.core._math", ["src/yaw/core/math.c"], include_dirs=[np.get_include()]
)


if __name__ == "__main__":
    setup(
        ext_modules=[
            ext_module_math,
        ],
        url="https://github.com/jlvdb/yet_another_wizz.git",
    )



"""This module implements the default values for the different configuration
objects. Each of the ``*Config`` objects has a corresponding class holding just
the default values, as listed below.
"""

__all__ = [
    "NotSet",
    "Scales",
    "AutoBinning",
    "Backend",
    "Configuration",
    "Resampling",
    "backend",
]


class _NotSet_meta(type):
    def __repr__(self) -> str:
        return "NotSet"  # pragma: no cover

    def __bool__(self) -> bool:
        return False


class NotSet(metaclass=_NotSet_meta):
    pass


class Scales:
    rweight = None
    rbin_num = 50


class AutoBinning:
    method = "linear"
    zbin_num = 30


class Backend:
    thread_num = None
    crosspatch = True
    rbin_slop = 0.01


class Configuration:
    scales = Scales
    binning = AutoBinning
    backend = Backend
    cosmology = "Planck15"


class Resampling:
    method = "jackknife"
    crosspatch = True
    n_boot = 500
    global_norm = False
    seed = 12345


backend = "scipy"



"""This module implements the catalog class. Catalogs hold all necessary data,
manage file loading, construction of spatial patches, and implement the pair
counting needed for the correlation function measurements.

The most important member is the catalog factory class :obj:`NewCatalog`,
which provides an interface to create new catalog instances for each of the
supported correlation measurement backends. The :obj:`BaseCatalog` defines the
common catalogue interface and must be subclasses by all other backend
implementations.
"""

# make backends available and make sure they are registered
from yaw.catalogs.catalog import BaseCatalog
from yaw.catalogs.factory import NewCatalog
from yaw.catalogs.linkage import PatchLinkage

# isort: split
from yaw.catalogs import scipy, treecorr  # noqa

__all__ = ["BaseCatalog", "NewCatalog", "PatchLinkage"]



"""This module implements various small utility functions, mostly for string
formatting numbers.
"""

from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import TypeVar, Union

import numpy as np
import tqdm
from numpy.typing import NDArray

__all__ = [
    "job_progress_bar",
    "LimitTracker",
    "long_num_format",
    "bytes_format",
    "format_float_fixed_width",
]


TypePathStr = Union[Path, str]
Tjob = TypeVar("Tjob")


def job_progress_bar(
    iterable: Iterable[Tjob], total: int | None = None
) -> Iterable[Tjob]:
    """Configure and return a tqdm progress bar with custom format."""
    config = dict(delay=0.5, leave=False, smoothing=0.1, unit="jobs")
    return tqdm.tqdm(iterable, total=total, **config)


class LimitTracker:
    """Tracks the global minimum and maximum of batches of arrays."""

    def __init__(self):
        self.min = +np.inf
        self.max = -np.inf

    def update(self, data: NDArray | None):
        """Process a new array with data and update the limits if necessary."""
        if data is not None:
            self.min = np.minimum(self.min, np.min(data))
            self.max = np.maximum(self.max, np.max(data))

    def get(self):
        """Return the minimum and maximum values processed so far."""
        vmin = None if np.isinf(self.min) else self.min
        vmax = None if np.isinf(self.max) else self.max
        return vmin, vmax


def long_num_format(x: float) -> str:
    """Format a floating point number as string with a numerical suffix.

    E.g.: 1234.0 is converted to ``1.24K``.
    """
    x = float(f"{x:.3g}")
    exp = 0
    while abs(x) >= 1000:
        exp += 1
        x /= 1000.0
    prefix = str(x).rstrip("0").rstrip(".")
    suffix = ["", "K", "M", "B", "T"][exp]
    return prefix + suffix


def bytes_format(x: float) -> str:
    """Format a floating point number as string in bytes notation.

    E.g.: 123.0 is converted to ``123B``, 1234.0 is converted to ``1.23KB``.
    """
    x = float(f"{x:.3g}")
    exp = 0
    while abs(x) >= 1024:
        exp += 1
        x /= 1024.0
    prefix = f"{x:.3f}"[:4].rstrip(".")
    suffix = ["B ", "KB", "MB", "GB", "TB"][exp]
    return prefix + suffix


def format_float_fixed_width(value: float, width: int) -> str:
    """Format a floating point number as string with fixed width."""
    string = f"{value: .{width}f}"[:width]
    if "nan" in string or "inf" in string:
        string = f"{string.strip():>{width}s}"
    return string



"""This module provides some precomputed example data, which are loaded when
importing the module, e.g.

>>> from yaw import examples  # reads the data sets from disk
>>> examples.w_sp
CorrFunc(n_bins=30, z='0.070...1.420', dd=True, dr=True, rd=False, rr=False, n_patches=64)
"""

from pathlib import Path

import yaw

__all__ = [
    "w_sp",
    "w_ss",
    "w_pp",
    "normalised_counts",
    "patched_count",
    "patched_total",
]


_path = Path(__file__).parent

w_sp = yaw.CorrFunc.from_file(_path / "cross_1.hdf")
"""Example data from a crosscorrelation measurement
(:obj:`~yaw.correlation.CorrFunc` instance)."""

w_ss = yaw.CorrFunc.from_file(_path / "auto_reference.hdf")
"""Example data from a reference sample autocorrelation measurement
(:obj:`~yaw.correlation.CorrFunc` instance)."""

w_pp = yaw.CorrFunc.from_file(_path / "auto_unknown_1.hdf")
"""Example data from a unknown sample autocorrelation measurement
(:obj:`~yaw.correlation.CorrFunc` instance)."""


normalised_counts = w_sp.dd
"""Example data for patch-wise, normalised pair counts
(:obj:`~yaw.correlation.paircounts.NormalisedCounts` instance,
from :obj:`w_sp.dd`)"""

patched_count = normalised_counts.count
"""Example data for patch-wise pair counts
(:obj:`~yaw.correlation.paircounts.PatchedCount` instance,
from :obj:`w_sp.dd.count`)"""

patched_total = normalised_counts.total
"""Example data for patch-wise total number of objects
(:obj:`~yaw.correlation.paircounts.PatchedTotal` instance,
from :obj:`w_sp.dd.total`)"""



"""This module implements a simple class to generate uniform randoms on a
rectangular footprint.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from yaw.core.logging import TimedLog
from yaw.core.parallel import POOL_SHARE, ParallelHelper, SharedArray
from yaw.core.utils import long_num_format

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray
    from pandas import DataFrame

    from yaw.catalogs import BaseCatalog

__all__ = ["UniformRandoms"]


logger = logging.getLogger(__name__)


class UniformRandoms:
    """Generator for uniform randoms on a rectangular footprint.

    Generates points uniform in right ascension and declination. Additional
    features can be cloned by sampling values from an external data catalogue
    (e.g. spectroscopic or photometric redshifts).

    Internally uses cylindrical coordinates :math:`(x, y)`, which are an equal
    area projection. Point are generated in cylindrical coordinates and
    transformed back to spherical coordinates :math:`(\\alpha, \\delta)`:

    :math:`\\alpha \\leftrightarrow x \\quad \\sin{\\delta} \\leftrightarrow y`
    """

    def __init__(
        self,
        ra_min: float,
        ra_max: float,
        dec_min: float,
        dec_max: float,
        seed: int = 12345,
    ) -> None:
        """Create a new generator for a the given footprint.

        Args:
            ra_min (:obj:`float`):
                Minimum right ascenion to generate, in degrees.
            ra_max (:obj:`float`):
                Maximum right ascenion to generate, in degrees.
            dec_min (:obj:`float`):
                Minimum declination to generate, in degrees.
            dec_max (:obj:`float`):
                Maximum declination to generate, in degrees.
            seed (:obj:`int`, optional):
                Seed to use for the random generator.
        """
        self.x_min, self.y_min = self.sky2cylinder(ra_min, dec_min)
        self.x_max, self.y_max = self.sky2cylinder(ra_max, dec_max)
        self.rng = np.random.SeedSequence(seed)

    @classmethod
    def from_catalog(cls, cat: BaseCatalog, seed: int = 12345) -> UniformRandoms:
        """Create a new generator with a rectangular footprint obtained from the
        coordinate range of a given data catalogue.

        Args:
            cat (:obj:`yaw.catalogs.BaseCatalog`):
                Catalog instance from which the right ascension and declination
                range is computed.
            seed (:obj:`int`, optional):
                Seed to use for the random generator.
        """
        return cls(
            np.rad2deg(cat.ra.min()),
            np.rad2deg(cat.ra.max()),
            np.rad2deg(cat.dec.min()),
            np.rad2deg(cat.dec.max()),
            seed=seed,
        )

    @staticmethod
    def sky2cylinder(
        ra: float | NDArray[np.float_], dec: float | NDArray[np.float_]
    ) -> NDArray:
        """Conversion from spherical to cylindrical coordinates.

        Args:
            ra (:obj:`float`, :obj:`NDArray`):
                Right ascension(s) to convert to cylindrical coordinates.
            dec (:obj:`float`, :obj:`NDArray`):
                Right ascension(s) to convert to cylindrical coordinates.

        Returns:
            :obj:`NDArray`:
                Array with of points in cylindrical coordinates of shape
                `(N, 2)`.
        """
        x = np.deg2rad(ra)
        y = np.sin(np.deg2rad(dec))
        return np.transpose([x, y])

    @staticmethod
    def cylinder2sky(
        x: float | NDArray[np.float_], y: float | NDArray[np.float_]
    ) -> float | NDArray[np.float_]:
        """Conversion from cylindrical to spherical coordinates.

        Args:
            x (:obj:`float`, :obj:`NDArray`):
                `x`-coordinate(s) to convert to spherical coordinates.
            y (:obj:`float`, :obj:`NDArray`):
                `y`-coordinate(s) to convert to spherical coordinates.

        Returns:
            :obj:`NDArray`:
                Array with of points in spherical coordinates of shape `(N, 2)`.
        """
        ra = np.rad2deg(x)
        dec = np.rad2deg(np.arcsin(y))
        return np.transpose([ra, dec])

    def generate(
        self,
        size: int,
        names: list[str, str] | None = None,
        draw_from: dict[str, NDArray] | None = None,
        n_threads: int = 1,
    ) -> DataFrame:
        """Generate new random points.

        Generate a specified number of points, additionally draw extra data
        features form a list of input values. Results are returned in a data
        frame.

        Args:
            size (:obj:`int`):
                Number of random points to generate.
            name (:obj:`tuple[str, str]`, optional):
                Name of the right ascension and declination columns in the
                output data frame. Default is ``ra`` and ``dec``.
            draw_from (:obj:`dict[str, NDArray]`, optional):
                Dictionary of data arrays. If provided, a random sample (with
                repetition) is drawn from these arrays and assigned to the
                output data frame. The dictionary keys are used to name the
                columns in the output.
            n_threads (:obj:`int`, optional):
                Generate data in parallel using subprocesses, default is
                parallel processing disabled.

                .. deprecated:: 2.3.2
                    No performance gain observed. May be removed in a future
                    version.

        Returns:
            :obj:`pandas.DataFrame`:
                Data frame with uniform random coordinates and optionally
                additional features draw from input data.
        """
        # seeds for threads
        if size <= 100 * n_threads:  # there is some kind of floor
            n_threads = 1
        seeds = self.rng.spawn(n_threads)
        # distribute load
        idx_break = np.linspace(0, size, n_threads + 1).astype(np.int_)
        start = idx_break[:-1]
        end = idx_break[1:]

        # create output arrays
        shares = dict(
            ra=SharedArray.empty((size,), "f8"), dec=SharedArray.empty((size,), "f8")
        )
        if draw_from is not None:
            shares["in"] = {
                key: SharedArray.from_numpy(array) for key, array in draw_from.items()
            }
            shares["out"] = {
                key: SharedArray.empty((size,), array.dtype)
                for key, array in draw_from.items()
            }
        if names is None:
            names = ["ra", "dec"]

        # process
        msg = f"generate ({long_num_format(size)} uniform randoms)"
        with TimedLog(logger.info, msg):
            with ParallelHelper(_generate_uniform_randoms, n_threads) as pool:
                pool.shares = shares  # assign in bulk
                pool.add_constant(self)
                pool.add_iterable(seeds)
                pool.add_iterable(start)
                pool.add_iterable(end)
                pool.result()  # output direclty pasted into shared arrays
                # convert to dataframe
                rand = pd.DataFrame(
                    {
                        names[0]: pool.shares["ra"].to_numpy(copy=True),
                        names[1]: pool.shares["dec"].to_numpy(copy=True),
                    }
                )
                if draw_from is not None:
                    for col, data in pool.shares["out"].items():
                        rand[col] = data.to_numpy(copy=True)
                # will delete shared arrays after this
        return rand


def _generate_uniform_randoms(
    inst: UniformRandoms, seed: np.random.SeedSequence, start: int, end: int
) -> int:
    rng = np.random.default_rng(seed)
    size = end - start
    ra = POOL_SHARE["ra"].to_numpy()
    dec = POOL_SHARE["dec"].to_numpy()
    # generate positions
    x = rng.uniform(inst.x_min, inst.x_max, size)
    y = rng.uniform(inst.y_min, inst.y_max, size)
    ra_dec = UniformRandoms.cylinder2sky(x, y)
    ra[start:end] = ra_dec[:, 0]
    dec[start:end] = ra_dec[:, 1]
    # generate random draw
    if "in" in POOL_SHARE:
        N = None
        for col, data in POOL_SHARE["in"].items():
            if N is None:
                if len(data.shape) > 1:
                    raise ValueError("data to draw from must be 1-dimensional")
                N = len(data)
            else:
                if len(data) != N:
                    raise ValueError("length of columns to draw from does not match")
        draw_idx = rng.integers(0, N, size=size)
        # draw and insert data
        for col, data in POOL_SHARE["in"].items():
            rand = POOL_SHARE["out"][col].to_numpy()
            rand[start:end] = data.to_numpy()[draw_idx]
    return size



"""
yaw
===

*yet_another_wizz* is a python package to efficiently compute cross-correlation
redshifts, also know as clustering redshifts and is hosted on github:

- code: https://github.com/jlvdb/yet_another_wizz.git
- docs: https://yet-another-wizz.readthedocs.io/

The method allows to estimate the unknown redshift distribution of a galaxy
sample by correlating the on-sky positions with a reference sample with known
redshifts. This implementation is based on the single bin correlation
measurement of the correlation amplitude, introduced by Schmidt et al. (2013,
`arXiv:1303.0292 <https://arxiv.org/abs/1303.0292>`_).

Author: Jan Luca van den Busch
        (Ruhr-Universität Bochum, Astronomisches Institut)
"""


import logging as _logging

_logging.getLogger(__name__).addHandler(_logging.NullHandler())  # noqa

from yaw.catalogs import NewCatalog
from yaw.config import Configuration, ResamplingConfig
from yaw.core.cosmology import Scale
from yaw.core.math import global_covariance
from yaw.correlation import CorrData, CorrFunc, autocorrelate, crosscorrelate
from yaw.deprecated import (
    CorrelationData,
    CorrelationFunction,
    HistogramData,
    PairCountResult,
)
from yaw.randoms import UniformRandoms
from yaw.redshifts import HistData, RedshiftData

__version__ = "2.5.1"

__all__ = [
    "NewCatalog",
    "Configuration",
    "ResamplingConfig",
    "Scale",
    "global_covariance",
    "CorrData",
    "CorrFunc",
    "autocorrelate",
    "crosscorrelate",
    "UniformRandoms",
    "HistData",
    "RedshiftData",
    # deprecated
    "CorrelationData",
    "CorrelationFunction",
    "HistogramData",
    "PairCountResult",
]



"""This module implements a special containers that compute and describe
redshift distributions. These containers provide methods for error and
covariance estimation, plotting and computing mean redshifts.

True redshift distributions can be expressed through the :obj:`HistData` class
and can be computed from the :meth:`~yaw.catalogs.BaseCatalog.true_redshifts`
method of :obj:`~yaw.catalogs.BaseCatalog`.

Clustering redshift estimates can be expressed through the :obj:`RedshiftData`
class. Its different constructor methods, :meth:`~RedshiftData.from_corrdata`
and :meth:`~RedshiftData.from_corrfuncs`, provide easy interfaces to compute the
clustering redshifts and to mitigate the evolving galaxy bias, if the
corresponding autocorrelation functions (e.g. of the reference sample) are
known.
"""

from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import scipy.optimize
from deprecated import deprecated

from yaw.config import ResamplingConfig
from yaw.core.containers import SampledValue
from yaw.core.logging import LogCustomWarning, TimedLog
from yaw.core.math import rebin, shift_histogram
from yaw.core.utils import TypePathStr
from yaw.correlation import CorrData

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

    from yaw.correlation import CorrFunc

__all__ = ["RedshiftData", "HistData"]


logger = logging.getLogger(__name__)


@dataclass(frozen=True, repr=False, eq=False)
class RedshiftData(CorrData):
    """Container class object for redshift estimates.

    Contains the redshift binning, estimated fraction of galaxies at the given
    redshift (**not** a density), and resampled fractions (e.g. from jackknife
    or bootstrap). The resampled values are used to compute error estimates and
    covariance/correlation matrices. Provides some plotting methods for
    convenience.

    This container holds data in the form of
    :math:`\\frac{w_\\rm{sp}(z)}{\\sqrt{\\Delta z \\, w_\\rm{ss}(z) \\, w_\\rm{pp}(z)}}`,
    where :math:`w_\\rm{sp}` is the crosscorrelation function, and
    :math:`w_\\rm{ss}` and :math:`w_\\rm{pp}` are autocorrelation functions that
    account for the evolving galaxy bias. If no autocorrelation is provided, the
    data is still scaled by :math:`1/\\Delta z` compared to the crosscorrelation
    data in :obj:`~yaw.correlation.CorrData`.

    .. Note::
        This container should be constructed from a crosscorrelation measurement
        with one of the preferred methods :meth:`from_corrdata` or
        :meth:`from_corrfuncs`. These additionally allow galaxy bias mitigation
        by specifying any additionally measured autocorrelation functions.

    The comparison, addition and subtraction and indexing rules are inherited
    from :obj:`~yaw.core.containers.SampledData`, check the examples there.

    .. rubric:: Examples

    Create a redshift estimate from a crosscorrelation function and correct for
    the evolving bias of the reference sample using its autocorrelation
    function:

    >>> from yaw.examples import w_sp  # crosscorrelation
    >>> from yaw.examples import w_ss  # reference sample autocorrelation
    >>> nz = yaw.yaw.RedshiftData.from_corrfuncs(w_sp, ref_corr=w_ss)
    RedshiftData(n_bins=30, z='0.070...1.420', n_samples=64, method='jackknife')

    Use a different estimator when sampling the autocorrelation function, e.g.
    the Peebles-Hauser estimator:

    >>> nz = yaw.RedshiftData.from_corrfuncs(w_sp, ref_corr=w_ss, ref_est="PH")
    RedshiftData(n_bins=30, z='0.070...1.420', n_samples=64, method='jackknife')

    View the data for a subset of the redshift bins:

    >>> nz.bins[5:9].data
    array([2.5234212 , 1.96617211, 1.05342   , 0.67866257])

    View the same subset as series:

    >>> nz.bins[5:9].get_data()
    (0.295, 0.34]    2.523421
    (0.34, 0.385]    1.966172
    (0.385, 0.43]    1.053420
    (0.43, 0.475]    0.678663
    dtype: float64

    Get the redshift bin centers for these bins:

    >>> nz.bins[5:9].mids
    array([0.3175, 0.3625, 0.4075, 0.4525])

    Plot the redshift distribution, indicating a zero-line

    >>> nz.plot(zero_line=True)
    <Axes: >

    .. figure:: ../../_static/ncc_example.png
        :width: 400
        :alt: example clustering redshfit estimate

    Args:
        binning (:obj:`pandas.IntervalIndex`):
            The redshift bin edges used for this correlation function.
        data (:obj:`NDArray`):
            The correlation function values.
        samples (:obj:`NDArray`):
            The resampled correlation function values.
        method (:obj:`str`):
            The resampling method used, see :class:`~yaw.ResamplingConfig` for
            available options.
        info (:obj:`str`, optional):
            Descriptive text included in the headers of output files produced
            by :func:`CorrData.to_files`.
    """

    @classmethod
    @deprecated(reason="renamed to RedshiftData.from_corrdata", version="2.3.2")
    def from_correlation_data(cls, *args, **kwargs):
        """
        .. deprecated:: 2.3.2
            Renamed to :meth:`from_corrdata`.
        """
        return cls.from_corrdata(*args, **kwargs)

    @classmethod
    def from_corrdata(
        cls,
        cross_data: CorrData,
        ref_data: CorrData | None = None,
        unk_data: CorrData | None = None,
        info: str | None = None,
    ) -> RedshiftData:
        """Compute redshift estimate from readily sampled function data.

        The required argument is a crosscorrelation measurement, additional
        parameters can specify sample autocorrelation measurements that are used
        to mitigate the evolving galaxy bias.

        Args:
            cross_corr (:obj:`CorrData`):
                Data from the sampled cross-correlation function.
            ref_corr (:obj:`CorrData`, optional):
                Data from the sampled reference sample autocorrelation function.
                Used to mitigate reference bias evolution.
            unk_corr (:obj:`CorrData`, optional):
                Data from the sampled unknown sample autocorrelation function.
                Used to mitigate unknown bias evolution.

        Returns:
            :obj:`RedshiftData`
        """
        logger.debug("computing clustering redshifts from correlation function samples")
        w_sp_data = cross_data.data
        w_sp_samp = cross_data.samples
        mitigate = []

        if ref_data is None:
            w_ss_data = np.float64(1.0)
            w_ss_samp = np.float64(1.0)
        else:
            try:
                ref_data.is_compatible(cross_data, require=True)
            except ValueError as e:
                raise ValueError(
                    "'ref_corr' correlation data is not compatible with 'cross_data'"
                ) from e
            w_ss_data = ref_data.data
            w_ss_samp = ref_data.samples
            mitigate.append("reference")

        if unk_data is None:
            w_pp_data = np.float64(1.0)
            w_pp_samp = np.float64(1.0)
        else:
            try:
                unk_data.is_compatible(cross_data, require=True)
            except ValueError as e:
                raise ValueError(
                    "'unk_data' correlation data is not compatible with " "'cross_data'"
                ) from e
            w_pp_data = unk_data.data
            w_pp_samp = unk_data.samples
            mitigate.append("unknown")

        if len(mitigate) > 0:
            logger.debug(f"mitigating {' and '.join(mitigate)} sample bias")
        N = cross_data.n_samples
        dzsq_data = cross_data.dz**2
        dzsq_samp = np.tile(dzsq_data, N).reshape((N, -1))
        with LogCustomWarning(
            logger, "invalid values encountered in redshift estimate"
        ):
            nz_data = w_sp_data / np.sqrt(dzsq_data * w_ss_data * w_pp_data)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            nz_samp = w_sp_samp / np.sqrt(dzsq_samp * w_ss_samp * w_pp_samp)
        return cls(
            binning=cross_data.binning,
            data=nz_data,
            samples=nz_samp,
            method=cross_data.method,
            info=info,
        )

    @classmethod
    @deprecated(reason="renamed to RedshiftData.from_corrfuncs", version="2.3.2")
    def from_correlation_functions(cls, *args, **kwargs):
        """
        .. deprecated:: 2.3.2
            Renamed to :meth:`from_corrfuncs`.
        """
        return cls.from_corrfuncs(*args, **kwargs)

    @classmethod
    def from_corrfuncs(
        cls,
        cross_corr: CorrFunc,
        ref_corr: CorrFunc | None = None,
        unk_corr: CorrFunc | None = None,
        *,
        cross_est: str | None = None,
        ref_est: str | None = None,
        unk_est: str | None = None,
        config: ResamplingConfig | None = None,
        info: str | None = None,
    ) -> RedshiftData:
        """Sample correlation functions to compute a redshift estimate.

        The required argument is a crosscorrelation measurement, additional
        parameters can specify sample autocorrelation measurements that are used
        to mitigate the evolving galaxy bias.

        Args:
            cross_corr (:obj:`CorrFunc`):
                The measured cross-correlation function.
            ref_corr (:obj:`CorrFunc`, optional):
                The measured reference sample autocorrelation function. Used to
                mitigate reference bias evolution.
            unk_corr (:obj:`CorrFunc`, optional):
                The measured unknown sample autocorrelation function. Used to
                mitigate unknown bias evolution.

        Returns:
            :obj:`RedshiftData`
        """
        if config is None:
            config = ResamplingConfig()
        with TimedLog(
            logger.debug,
            f"estimating clustering redshifts with method '{config.method}'",
        ):
            # check compatibilty before sampling anything
            if ref_corr is not None:
                try:
                    cross_corr.is_compatible(ref_corr, require=True)
                except ValueError as e:
                    raise ValueError(
                        "'ref_corr' correlation function is not compatible "
                        "with 'cross_corr'"
                    ) from e
            if unk_corr is not None:
                try:
                    cross_corr.is_compatible(unk_corr, require=True)
                except ValueError as e:
                    raise ValueError(
                        "'unk_corr' correlation function is not compatible "
                        "with 'cross_corr'"
                    ) from e
            # sample pair counts and evaluate estimator
            cross_data = cross_corr.sample(config, estimator=cross_est)
            if ref_corr is not None:
                ref_data = ref_corr.sample(config, estimator=ref_est)
            else:
                ref_data = None
            if unk_corr is not None:
                unk_data = unk_corr.sample(config, estimator=unk_est)
            else:
                unk_data = None
            return cls.from_correlation_data(
                cross_data=cross_data, ref_data=ref_data, unk_data=unk_data, info=info
            )

    @property
    def _dat_desc(self) -> str:
        return "# n(z) estimate with symmetric 68% percentile confidence"

    @property
    def _smp_desc(self) -> str:
        return f"# {self.n_samples} {self.method} n(z) samples"

    @property
    def _cov_desc(self) -> str:
        return f"# n(z) estimate covariance matrix ({self.n_bins}x{self.n_bins})"

    def normalised(self, to: CorrData | None = None) -> RedshiftData:
        """Obtain a normalised copy of the data.

        Either attempts to normalise the data by integration along the redshift
        axis or by fitting it to a provided reference data container (e.g. a
        known redshift distribution in a :obj:`HistData` container).

        .. Note::
            The fit does not use the uncertainties but weights the data points
            inversely to their amplitude.

        Args:
            to (:obj:`CorrData`, optional):
                Reference data to which the stored values are normalised by
                fitting.

        Returns:
            :obj:`RedshiftData`
        """
        if to is None:
            norm = np.nansum(self.dz * self.data)
        else:
            y_from = self.data
            y_to = to.data
            mask = np.isfinite(y_from) & np.isfinite(y_to) & (y_to > 0.0)
            norm = scipy.optimize.curve_fit(
                lambda x, norm: y_from[mask] / norm,  # x is a dummy variable
                xdata=to.mids[mask],
                ydata=y_to[mask],
                p0=[1.0],
                sigma=1 / y_to[mask],
            )[0][0]
        return self.__class__(
            binning=self.get_binning(),
            data=self.data / norm,
            samples=self.samples / norm,
            method=self.method,
            info=self.info,
        )

    def mean(self):
        """Attempts to compute a mean redshift.

        .. Warning::
            This should be just considered an estimate since the redshift
            estimate is not a true probability density, due to residual negative
            correlation amplitudes.

        Returns:
            :obj:`~yaw.core.SampledValue`:
                Mean redshift for the redshift data and its samples in a data
                container.
        """
        norm = np.nansum(self.data)
        mean = np.nansum(self.data * self.mids) / norm
        samples = np.nansum(self.samples * self.mids, axis=1) / norm
        return SampledValue(value=mean, samples=samples, method=self.method)

    def rebin(self, bins: NDArray) -> RedshiftData:
        """Attempts recomute the data for a different redshift binning.

        .. Warning::
            The result may be inaccurate since the redshift estimate is not a
            true probability density, due to residual negative correlation
            amplitudes.

        Args:
            bins (:obj:`NDArray`):
                Edges of the new redshift bins. May exceed or cover just a
                fraction of the original redshift range.

        Returns:
            :obj:`RedshiftData`:
        """
        old_bins = self.edges
        # shift main values
        data = rebin(bins, old_bins, self.data)
        # shift the value samples
        samples = np.empty([self.n_samples, len(bins) - 1], data.dtype)
        for i, sample in enumerate(self.samples):
            samples[i] = rebin(bins, old_bins, sample)

        return self.__class__(
            binning=pd.IntervalIndex.from_breaks(bins),
            data=data,
            samples=samples,
            method=self.method,
            info=self.info,
        )

    def shift(
        self, dz: float | SampledValue = 0.0, *, amplitude: float | SampledValue = 1.0
    ) -> RedshiftData:
        """Attempts shift the data along the redshift axis.

        The shifting is performed by recomputing the redshift estimate with its
        original redshift bins which are shifted by some amount.

        .. Warning::
            The result may be inaccurate since the redshift estimate is not a
            true probability density, due to residual negative correlation
            amplitudes.

        Args:
            dz (:obj:`SampledValue` or :obj:`float`):
                The amplitude of the shift along the redshift axis. If the input
                provides samples of the shift, each redshift estimate sample is
                shifted individually to obtain a more accurate error estimate.
            amplitude (:obj:`SampledValue` or :obj:`float`):
                An optional ampltude factor applied to the redshift estimate.
                Same rules as for the ``dz`` parameter.

        Returns:
            :obj:`RedshiftData`:
        """
        if isinstance(amplitude, SampledValue):
            A_samples = amplitude.samples
            amplitude = amplitude.value
        else:
            A_samples = [amplitude] * self.n_samples
        if isinstance(dz, SampledValue):
            dz_samples = dz.samples
            dz = dz.value
        else:
            dz_samples = [dz] * self.n_samples

        bins = self.edges
        # shift main values
        data = shift_histogram(bins, self.data, A=amplitude, dx=dz)
        # shift the value samples
        samples = np.empty_like(self.samples)
        iter_samples = zip(self.samples, A_samples, dz_samples)
        for i, (sample, amplitude, dz) in enumerate(iter_samples):
            samples[i] = shift_histogram(bins, sample, A=amplitude, dx=dz)

        return self.__class__(
            binning=self.get_binning(),
            data=data,
            samples=samples,
            method=self.method,
            info=self.info,
        )


@dataclass(frozen=True, repr=False, eq=False)
class HistData(RedshiftData):
    """Container for histogram data.

    Contains the redshift binning, histogram counts, and resampled counts (e.g.
    from jackknife or bootstrap). The resampled values are used to compute error
    estimates and covariance/correlation matrices. Provides some plotting
    methods for convenience.

    Args:
        binning (:obj:`pandas.IntervalIndex`):
            The redshift bin edges used for this correlation function.
        data (:obj:`NDArray`):
            The correlation function values.
        samples (:obj:`NDArray`):
            The resampled correlation function values.
        method (:obj:`str`):
            The resampling method used, see :class:`~yaw.ResamplingConfig` for
            available options.
        info (:obj:`str`, optional):
            Descriptive text included in the headers of output files produced
            by :func:`CorrData.to_files`.
        density (:obj:`bool`):
            Whether the data is normalised, i.e. a density estimate.
    """

    density: bool = field(default=False)
    """Whether the data is normalised, i.e. a density estimate."""

    @property
    def _dat_desc(self) -> str:
        n = "normalised " if self.density else " "
        return f"# n(z) {n}histogram with symmetric 68% percentile confidence"

    @property
    def _smp_desc(self) -> str:
        n = "normalised " if self.density else " "
        return f"# {self.n_samples} {self.method} n(z) {n}histogram samples"

    @property
    def _cov_desc(self) -> str:
        n = "normalised " if self.density else " "
        return f"# n(z) {n}histogram covariance matrix ({self.n_bins}x{self.n_bins})"

    @classmethod
    def from_files(cls, path_prefix: TypePathStr) -> HistData:
        new = super().from_files(path_prefix)
        with open(f"{path_prefix}.dat") as f:
            line = f.readline()
            density = "normalised" in line
        return cls(
            binning=new.get_binning(),
            data=new.data,
            samples=new.samples,
            method=new.method,
            density=density,
        )

    def normalised(self, *args, **kwargs) -> HistData:
        """Obtain a normalised copy of the data.

        Normalises the data by integration along the redshift axis. This sets
        the containers :obj:`density` flag to ``True``.

        Returns:
            :obj:`HistData`
        """
        if self.density:  # guard from repeatedly altering the data
            return self
        zmin, zmax = self.edges[[0, -1]]
        width_correction = (zmax - zmin) / (self.n_bins * self.dz)
        data = self.data * width_correction
        samples = self.samples * width_correction
        norm = np.nansum(self.dz * data)
        return self.__class__(
            binning=self.get_binning(),
            data=data / norm,
            samples=samples / norm,
            method=self.method,
            info=self.info,
            density=True,
        )

    def mean(self) -> SampledValue:
        """Compute the mean redshift.

        Returns:
            :obj:`~yaw.core.SampledValue`:
                Mean redshift for the redshift data and its samples in a data
                container.
        """
        normed = self.normalised()
        norm = np.nansum(normed.data)
        mean = np.nansum(normed.data * normed.mids) / norm
        samples = np.nansum(normed.samples * normed.mids, axis=1) / norm
        return SampledValue(value=mean, samples=samples, method=normed.method)

    def rebin(self, bins: NDArray) -> HistData:
        """Recomute the data for a different redshift binning.

        .. Warning::
            The result may be inaccurate since the result is interpolated
            step-wise.

        Args:
            bins (:obj:`NDArray`):
                Edges of the new redshift bins. May exceed or cover just a
                fraction of the original redshift range.

        Returns:
            :obj:`HistData`:
        """
        result = super().rebin(bins)
        object.__setattr__(self, "density", self.density)
        return result

    def shift(
        self, dz: float | SampledValue = 0.0, *, amplitude: float | SampledValue = 1.0
    ) -> HistData:
        """Shifts the data along the redshift axis.

        The shifting is performed by recomputing the histogram with its original
        redshift bins which are shifted by some amount.

        .. Warning::
            The result may be inaccurate since the result is interpolated
            step-wise.

        Args:
            dz (:obj:`SampledValue` or :obj:`float`):
                The amplitude of the shift along the redshift axis. If the input
                provides samples of the shift, each redshift estimate sample is
                shifted individually to obtain a more accurate error estimate.
            amplitude (:obj:`SampledValue` or :obj:`float`):
                An optional ampltude factor applied to the redshift estimate.
                Same rules as for the ``dz`` parameter.

        Returns:
            :obj:`HistData`:
        """
        result = super().shift(dz, amplitude=amplitude)
        if amplitude == 1.0:
            object.__setattr__(self, "density", self.density)
        else:
            object.__setattr__(self, "density", False)
        return result
